from typing import Optional

from pydantic import BaseModel

from qualibrate_runner.core.models.last_run import RunError, RunStatus


class WorkflowStatus(BaseModel):
    active: bool
    status: RunStatus
    nodes_completed: int
    nodes_total: int
    run_duration: float
    error: Optional[RunError] = None



from typing import Any

from pydantic import BaseModel, Field


class Reference(BaseModel):
    config_path: str
    reference_path: str
    index_start: int
    index_end: int
    value: Any = None
    solved: bool = False


class PathWithSolvingReferences(BaseModel):
    config_path: str
    value: Any = None
    solved: bool = False
    references: list[Reference] = Field(default_factory=list)



import os
import sys
from pathlib import Path
from typing import Any, Mapping

import click
import tomli_w
from click.core import ParameterSource

from qualibrate_runner.config import (
    CONFIG_KEY as QUALIBRATE_RUNNER_CONFIG_KEY,
)
from qualibrate_runner.config import (
    DEFAULT_CONFIG_FILENAME,
    QUALIBRATE_PATH,
    QualibrateRunnerSettings,
    get_config_file,
)
from qualibrate_runner.config.validation import get_config_model_or_print_error

if sys.version_info[:2] < (3, 11):
    import tomli as tomllib
else:
    import tomllib

__all__ = ["config_command"]


def not_default(ctx: click.Context, arg_key: str) -> bool:
    return ctx.get_parameter_source(arg_key) in (
        ParameterSource.COMMANDLINE,
        ParameterSource.ENVIRONMENT,
    )


def get_config(config_path: Path) -> tuple[dict[str, Any], Path]:
    """Returns config and path to file"""
    config_file = get_config_file(config_path, raise_not_exists=False)
    if config_file.is_file():
        return tomllib.loads(config_file.read_text()), config_path
    return {}, config_file


def _config_from_sources(
    ctx: click.Context, from_file: dict[str, Any]
) -> dict[str, Any]:
    config_keys = QualibrateRunnerSettings.model_fields.keys()
    runner_mapping: dict[str, str] = {k: k for k in config_keys}
    for arg_key, arg_value in ctx.params.items():
        not_default_arg = not_default(ctx, arg_key)
        if arg_key in runner_mapping.keys():
            if not_default_arg or runner_mapping[arg_key] not in from_file:
                from_file[runner_mapping[arg_key]] = arg_value
    return from_file


def write_config(
    config_file: Path,
    common_config: dict[str, Any],
    qrs: QualibrateRunnerSettings,
    confirm: bool = True,
) -> None:
    exported_data = qrs.model_dump(mode="json")
    if confirm:
        _confirm(config_file, exported_data)
    if not config_file.parent.exists():
        config_file.parent.mkdir(parents=True)
    common_config[QUALIBRATE_RUNNER_CONFIG_KEY] = exported_data
    with config_file.open("wb") as f_out:
        tomli_w.dump(common_config, f_out)


def _print_config(data: Mapping[str, Any], depth: int = 0) -> None:
    max_key_len = max(map(len, map(str, data.keys())))
    click.echo(
        os.linesep.join(
            f"{' ' * 4 * depth}{f'{k} :':<{max_key_len + 3}} {v}"
            for k, v in data.items()
            if not isinstance(v, Mapping)
        )
    )
    mappings = filter(lambda x: isinstance(x[1], Mapping), data.items())
    for mapping_k, mapping_v in mappings:
        click.echo(f"{' ' * 4 * depth}{mapping_k} :")
        _print_config(mapping_v, depth + 1)


def _confirm(config_file: Path, exported_data: dict[str, Any]) -> None:
    click.echo(f"Config file path: {config_file}")
    click.echo(click.style("Generated config:", bold=True))
    _print_config(exported_data)
    confirmed = click.confirm("Do you confirm config?", default=True)
    if not confirmed:
        click.echo(
            click.style(
                (
                    "The configuration has not been confirmed. "
                    "Rerun config script."
                ),
                fg="yellow",
            )
        )
        exit(1)


@click.command(name="config")
@click.option(
    "--config-path",
    type=click.Path(
        exists=False,
        path_type=Path,
    ),
    default=QUALIBRATE_PATH / DEFAULT_CONFIG_FILENAME,
    show_default=True,
)
@click.option(
    "--auto-accept",
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--overwrite",
    type=bool,
    default=False,
    is_flag=True,
    help="Ignore existing config and force overwrite values",
)
@click.option(
    "--calibration-library-resolver",
    type=click.STRING,
    default="qualibrate.QualibrationLibrary",
)
@click.option(
    "--calibration-library-folder",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    default=QUALIBRATE_PATH / "calibrations",
)
@click.pass_context
def config_command(
    ctx: click.Context,
    config_path: Path,
    auto_accept: bool,
    overwrite: bool,
    calibration_library_resolver: str,
    calibration_library_folder: Path,
) -> None:
    common_config, config_file = get_config(config_path)
    runner_config = (
        common_config.get(QUALIBRATE_RUNNER_CONFIG_KEY, {})
        if not overwrite
        else {}
    )
    runner_config = _config_from_sources(ctx, runner_config)

    qrs = get_config_model_or_print_error(runner_config)
    if qrs is None:
        return
    write_config(config_file, common_config, qrs, confirm=not auto_accept)



from collections import defaultdict
from typing import Any, List, Mapping, Optional, Sequence, Set, Tuple, cast

import jsonpatch
import jsonpointer

from qualibrate_runner.config.references.models import (
    PathWithSolvingReferences,
    Reference,
)

TEMPLATE_START = "${#"


def find_all_references(
    document: Mapping[str, Any], current_path: Optional[list[str]] = None
) -> Sequence[Reference]:
    if current_path is None:
        current_path = []
    to_resolve: list[Reference] = []
    for key, value in document.items():
        if isinstance(value, Mapping):
            to_resolve.extend(find_all_references(value, current_path + [key]))
        elif isinstance(value, str):
            config_path = "/" + "/".join(current_path + [key])
            template_start_index = value.find(TEMPLATE_START)
            while template_start_index != -1:
                template_end_index = value.find("}", template_start_index)
                if template_end_index == -1:
                    break
                to_resolve.append(
                    Reference(
                        config_path=config_path,
                        reference_path=value[
                            template_start_index + 3 : template_end_index
                        ].strip(),
                        index_start=template_start_index,
                        index_end=template_end_index,
                    )
                )
                template_start_index = value.find(
                    TEMPLATE_START, template_end_index + 1
                )
    return to_resolve


def check_cycles_in_references(
    references: Mapping[str, Sequence[str]],
) -> Tuple[bool, Optional[Sequence[str]]]:
    """Return True if the references has a cycle.

    >>> check_cycles_in_references({"a": ("b",), "b": ("c",), "c": ("a",)})
    (True, ['a', 'b', 'c', 'a'])
    >>> check_cycles_in_references({"a": ("b",), "b": ("c",), "c": ("d",)})
    (False, None)

    """
    path: List[str] = []
    visited: Set[str] = set()
    cycled_item: str = ""

    def visit(vertex: str) -> bool:
        nonlocal cycled_item
        if vertex in visited:
            return False
        visited.add(vertex)
        path.append(vertex)
        for neighbour in references.get(vertex, ()):
            if neighbour in path or visit(neighbour):
                cycled_item = vertex
                return True
        path.pop()  # == path.remove(vertex):
        return False

    if any(visit(v) for v in references):
        return True, [*path, cycled_item]
    return False, None


def _resolve_references(
    path: str,
    path_with_references: dict[str, PathWithSolvingReferences],
    original_config: Mapping[str, Any],
    solved_references: dict[str, Any],
) -> None:
    if path_with_references[path].solved:
        # path already solved
        return
    config_item = path_with_references[path]
    references: list[Reference] = config_item.references
    not_solved_refs = filter(lambda ref: not ref.solved, references)
    for ref in not_solved_refs:
        if ref.reference_path in solved_references:
            ref.value = solved_references[ref.reference_path]
            ref.solved = True
        elif ref.reference_path in path_with_references:
            _resolve_references(
                ref.reference_path,
                path_with_references,
                original_config,
                solved_references,
            )
            if ref.reference_path not in solved_references:
                raise ValueError(
                    f"Subreference '{ref.reference_path}' "
                    f"for '{ref.config_path}' not solved."
                )
            ref.value = solved_references[ref.reference_path]
            ref.solved = True
        else:
            value = jsonpointer.resolve_pointer(
                original_config, ref.reference_path, None
            )
            if value is None:
                raise ValueError(
                    f"Can't resolve reference item '{ref.reference_path}' "
                    f"for config path '{ref.config_path}'"
                )
            ref.value = value
            ref.solved = True
            solved_references[ref.reference_path] = value
    verify_all_solved = all(map(lambda ref: ref.solved, references))
    if not verify_all_solved:
        not_solved_refs = filter(lambda ref: not ref.solved, references)
        references_errors = (
            (
                f"- config path: '{ref.config_path}', "
                f"reference: '{ref.reference_path}';"
            )
            for ref in not_solved_refs
        )
        raise ValueError(
            "\n".join(
                [
                    "Some issues with solving references. Issued references:",
                    *references_errors,
                ]
            )
        )
    config_value = jsonpointer.resolve_pointer(
        original_config, config_item.config_path, None
    )
    if config_value is None or not isinstance(config_value, str):
        raise ValueError(
            f"Can't resolve config item '{config_item.config_path}'"
        )
    for ref in sorted(
        references, key=lambda ref: ref.index_start, reverse=True
    ):
        config_value = (
            f"{config_value[:ref.index_start]}"
            f"{ref.value}{config_value[ref.index_end + 1:]}"
        )
    config_item.value = config_value
    config_item.solved = True
    solved_references[config_item.config_path] = config_value


def resolve_references(config: dict[str, Any]) -> dict[str, Any]:
    references = find_all_references(config)
    references_seq = defaultdict(list)
    for reference in references:
        references_seq[reference.config_path].append(reference.reference_path)
    has_cycles, cycle = check_cycles_in_references(references_seq)
    if has_cycles:
        raise ValueError(f"Config contains cycle: {cycle}")
    solved_references: dict[str, Any] = {}
    path_with_references: dict[str, PathWithSolvingReferences] = {}
    for reference in references:
        path_with_refs = path_with_references.setdefault(
            reference.config_path,
            PathWithSolvingReferences(config_path=reference.config_path),
        )
        path_with_refs.references.append(reference)
    for path in path_with_references:
        _resolve_references(
            path,
            path_with_references,
            config,
            solved_references,
        )
    patches = [
        {"op": "replace", "path": path.config_path, "value": path.value}
        for path in path_with_references.values()
    ]
    return cast(dict[str, Any], jsonpatch.apply_patch(config, patches))


if __name__ == "__main__":
    _config = {
        "qualibrate": {"project": "my_project"},
        # "qualibrate": {"project": "${#/data_handler/project}"},
        "data_handler": {
            "root_data_folder": "/data/${#/data_handler/project}/subpath",
            "project": "${#/qualibrate/project}",
            # "project": "${#/qualibrate/project_}",
        },
    }
    res = resolve_references(_config)
    print(_config)
    print(res)



from datetime import datetime
from enum import Enum
from typing import Any, Mapping, Optional, Union

from pydantic import BaseModel, Field, computed_field
from qualibrate.models.run_summary.graph import GraphRunSummary
from qualibrate.models.run_summary.node import NodeRunSummary


class RunStatus(Enum):
    RUNNING = "running"
    FINISHED = "finished"
    ERROR = "error"


class RunnableType(Enum):
    NODE = "node"
    GRAPH = "graph"


class RunError(BaseModel):
    error_class: str
    message: str
    traceback: list[str]


class StateUpdate(BaseModel):
    key: str
    attr: Union[str, int]
    old: Any
    new: Any
    updated: bool = False


class LastRun(BaseModel):
    status: RunStatus
    started_at: datetime
    completed_at: Optional[datetime] = None
    name: str
    idx: int
    runnable_type: RunnableType
    passed_parameters: Mapping[str, Any] = Field(default_factory=dict)
    run_result: Optional[Union[NodeRunSummary, GraphRunSummary]] = None
    state_updates: Union[
        Mapping[str, StateUpdate], Mapping[str, Mapping[str, StateUpdate]]
    ] = Field(default_factory=dict)
    error: Optional[RunError] = None

    @computed_field
    def run_duration(self) -> float:
        duration = (
            self.completed_at - self.started_at
            if self.completed_at is not None
            else datetime.now() - self.started_at
        )
        return round(duration.total_seconds(), 3)



from pathlib import Path

CONFIG_KEY = "qualibrate_runner"
QUALIBRATE_PATH = Path().home() / ".qualibrate"
DEFAULT_CONFIG_FILENAME = "config.toml"
DEFAULT_QUALIBRATE_CONFIG_FILENAME = "qualibrate-runner.toml"
CONFIG_PATH_ENV_NAME = "QUALIBRATE_RUNNER_CONFIG_FILE"



from typing import Annotated, Any, Mapping, Optional, cast

from fastapi import APIRouter, Depends
from qualibrate.models.execution_history import ExecutionHistory
from qualibrate.qualibration_graph import QualibrationGraph

from qualibrate_runner.api.dependencies import get_state
from qualibrate_runner.config import State
from qualibrate_runner.core.models.last_run import LastRun, RunStatus
from qualibrate_runner.core.models.workflow import WorkflowStatus

last_run_router = APIRouter(prefix="/last_run")


@last_run_router.get("/")
def get_last_run(
    state: Annotated[State, Depends(get_state)],
) -> Optional[LastRun]:
    return state.last_run


@last_run_router.get("/workflow/status")
def get_workflow_status(
    state: Annotated[State, Depends(get_state)],
) -> Optional[WorkflowStatus]:
    if not isinstance(state.run_item, QualibrationGraph):
        return None
    graph: QualibrationGraph = state.run_item
    last_run = state.last_run
    run_duration = float(
        last_run.run_duration if last_run else 0.0  # type: ignore
    )
    return WorkflowStatus(
        status=last_run.status if last_run else RunStatus.FINISHED,
        active=state.is_running,
        nodes_completed=graph.completed_count(),
        nodes_total=len(graph._nodes),
        run_duration=run_duration,
        error=last_run.error if last_run else None,
    )


@last_run_router.get("/workflow/execution_history")
def get_execution_history(
    state: Annotated[State, Depends(get_state)],
    reverse: bool = False,
) -> Optional[Mapping[str, Any]]:
    if not isinstance(state.run_item, QualibrationGraph):
        return None
    graph: QualibrationGraph = state.run_item
    orch = graph._orchestrator
    if orch is None:
        raise RuntimeError("No graph orchestrator")
    history: ExecutionHistory = orch.get_execution_history()
    if reverse:
        history.items = list(reversed(history.items))
    return cast(
        Mapping[str, Any],
        history.model_dump(mode="json", serialize_as_any=True),
    )



import uvicorn
from fastapi import FastAPI

from qualibrate_runner.api.routes import base_router

app = FastAPI()
app.include_router(base_router)


def main(port: int, reload: bool) -> None:
    uvicorn.run("qualibrate_runner.app:app", port=port, reload=reload)


if __name__ == "__main__":
    main(port=8002, reload=False)



import os
from functools import lru_cache
from pathlib import Path
from typing import Annotated

from fastapi import Depends

from qualibrate_runner.config.file import get_config_file
from qualibrate_runner.config.models import QualibrateRunnerSettings
from qualibrate_runner.config.validation import (
    get_config_model_or_print_error,
    get_config_solved_references_or_print_error,
)
from qualibrate_runner.config.vars import (
    CONFIG_KEY,
    CONFIG_PATH_ENV_NAME,
)

__all__ = [
    "get_config_path",
    "get_settings",
]


@lru_cache
def get_config_path() -> Path:
    return get_config_file(os.environ.get(CONFIG_PATH_ENV_NAME))


@lru_cache
def get_settings(
    config_path: Annotated[Path, Depends(get_config_path)],
) -> QualibrateRunnerSettings:
    config = get_config_solved_references_or_print_error(config_path)
    if config is None:
        raise RuntimeError("Couldn't read config file")
    qrs = get_config_model_or_print_error(config.get(CONFIG_KEY, {}))
    if qrs is None:
        raise RuntimeError("Couldn't read config file")
    return qrs



from typing import Annotated, Mapping, Optional, cast

from fastapi import APIRouter, Depends, HTTPException, status

from qualibrate_runner.api.dependencies import (
    get_state,
)
from qualibrate_runner.config import (
    State,
)
from qualibrate_runner.core.models.last_run import (
    LastRun,
    RunStatus,
    StateUpdate,
)

others_router = APIRouter()


@others_router.get("/is_running")
def check_running(
    state: Annotated[State, Depends(get_state)],
) -> bool:
    return state.is_running


@others_router.post("/stop")
def stop_running(
    state: Annotated[State, Depends(get_state)],
    stop_graph_node: bool = False,
) -> bool:
    run_item = state.run_item
    if run_item is None:
        return False
    return bool(run_item.stop(stop_graph_node=stop_graph_node))


@others_router.post(
    "/record_state_update",
    description=(
        "Record that a state update entry belonging to the last run has been "
        "updated. This changed the state_updates entry to True  but does not "
        "update the snapshot."
    ),
)
def state_updated(
    state: Annotated[State, Depends(get_state)],
    key: str,
) -> Optional[LastRun]:
    if state.last_run is None or state.last_run.status != RunStatus.FINISHED:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="Node not executed or finished unsuccessful.",
        )
    state_updates = state.last_run.state_updates
    if key not in state_updates:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="Unknown state update key.",
        )
    state_updates = cast(Mapping[str, StateUpdate], state_updates)
    state_updates[key].updated = True
    return state.last_run



from fastapi import APIRouter

from .get_runnables import get_runnables_router
from .last_run import last_run_router
from .others import others_router
from .submit import submit_router

base_router = APIRouter()

base_router.include_router(submit_router)
base_router.include_router(get_runnables_router)
base_router.include_router(last_run_router)
base_router.include_router(others_router)



from .file import get_config_file, read_config_file
from .models import QualibrateRunnerSettings, State
from .resolvers import get_config_path, get_settings
from .vars import (
    CONFIG_KEY,
    CONFIG_PATH_ENV_NAME,
    DEFAULT_CONFIG_FILENAME,
    DEFAULT_QUALIBRATE_CONFIG_FILENAME,
    QUALIBRATE_PATH,
)

__all__ = [
    "CONFIG_KEY",
    "DEFAULT_CONFIG_FILENAME",
    "DEFAULT_QUALIBRATE_CONFIG_FILENAME",
    "QUALIBRATE_PATH",
    "CONFIG_PATH_ENV_NAME",
    "State",
    "QualibrateRunnerSettings",
    "get_config_file",
    "read_config_file",
    "get_config_path",
    "get_settings",
]



from functools import cache
from typing import Annotated, Mapping, cast

from fastapi import Depends, HTTPException
from qualibrate.qualibration_graph import QualibrationGraph
from qualibrate.qualibration_library import QualibrationLibrary
from qualibrate.qualibration_node import QualibrationNode

from qualibrate_runner.config import (
    QualibrateRunnerSettings,
    State,
    get_settings,
)


@cache
def get_state() -> State:
    return State()


@cache
def get_cached_library(
    settings: Annotated[QualibrateRunnerSettings, Depends(get_settings)],
) -> QualibrationLibrary:
    return settings.calibration_library_resolver(
        settings.calibration_library_folder
    )


def get_library(
    library: Annotated[QualibrationLibrary, Depends(get_cached_library)],
    rescan: bool = False,
) -> QualibrationLibrary:
    if rescan:
        library.rescan()
    return library


def get_nodes(
    library: Annotated[QualibrationLibrary, Depends(get_library)],
) -> Mapping[str, QualibrationNode]:
    return cast(Mapping[str, QualibrationNode], library.get_nodes())


def get_graphs(
    library: Annotated[QualibrationLibrary, Depends(get_library)],
) -> Mapping[str, QualibrationGraph]:
    return cast(Mapping[str, QualibrationGraph], library.get_graphs())


def get_node(
    name: str,
    nodes: Annotated[Mapping[str, QualibrationNode], Depends(get_nodes)],
) -> QualibrationNode:
    node = nodes.get(name)
    if node is None:
        raise HTTPException(status_code=422, detail=f"Unknown node name {name}")
    return node


def get_graph(
    name: str,
    graphs: Annotated[Mapping[str, QualibrationGraph], Depends(get_graphs)],
) -> QualibrationGraph:
    graph = graphs.get(name)
    if graph is None:
        raise HTTPException(
            status_code=422, detail=f"Unknown graph name {name}"
        )
    return graph



from typing import Annotated, Any, Mapping, Sequence, cast

from fastapi import APIRouter, Depends
from qualibrate.qualibration_graph import QualibrationGraph
from qualibrate.qualibration_node import QualibrationNode

from qualibrate_runner.api.dependencies import (
    get_graph as get_qgraph,
)
from qualibrate_runner.api.dependencies import (
    get_graphs as get_qgraphs,
)
from qualibrate_runner.api.dependencies import (
    get_node as get_qnode,
)
from qualibrate_runner.api.dependencies import (
    get_nodes as get_qnodes,
)

get_runnables_router = APIRouter()


@get_runnables_router.get("/get_nodes")
def get_nodes(
    nodes: Annotated[Mapping[str, QualibrationNode], Depends(get_qnodes)],
) -> Mapping[str, Any]:
    return {
        node_name: node.serialize(exclude_targets=True)
        for node_name, node in nodes.items()
    }


@get_runnables_router.get("/get_graphs")
def get_graphs(
    graphs: Annotated[Mapping[str, QualibrationNode], Depends(get_qgraphs)],
    cytoscape: bool = False,
) -> Mapping[str, Any]:
    return {
        graph_name: graph.serialize(cytoscape=cytoscape)
        for graph_name, graph in graphs.items()
    }


@get_runnables_router.get("/get_node")
def get_node(
    node: Annotated[QualibrationNode, Depends(get_qnode)],
) -> Mapping[str, Any]:
    return cast(Mapping[str, Any], node.serialize(exclude_targets=True))


@get_runnables_router.get("/get_graph")
def get_graph(
    graph: Annotated[QualibrationGraph, Depends(get_qgraph)],
    cytoscape: bool = False,
) -> Mapping[str, Any]:
    return cast(Mapping[str, Any], graph.serialize(cytoscape=cytoscape))


@get_runnables_router.get("/get_graph/cytoscape")
def get_graph_cytoscape(
    graph: Annotated[QualibrationGraph, Depends(get_qgraph)],
) -> Sequence[Mapping[str, Any]]:
    return cast(
        Sequence[Mapping[str, Any]],
        graph.cytoscape_representation(graph.serialize()),
    )



from pathlib import Path
from typing import Callable, Optional, Union

from pydantic import BaseModel, ConfigDict, DirectoryPath, ImportString
from pydantic_settings import BaseSettings, SettingsConfigDict
from qualibrate.qualibration_graph import QualibrationGraph
from qualibrate.qualibration_library import QualibrationLibrary
from qualibrate.qualibration_node import QualibrationNode

from qualibrate_runner.core.models.last_run import LastRun, RunStatus


class State(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    last_run: Optional[LastRun] = None
    run_item: Optional[Union[QualibrationNode, QualibrationGraph]] = None

    @property
    def is_running(self) -> bool:
        return (
            self.last_run is not None
            and self.last_run.status == RunStatus.RUNNING
        )


class QualibrateRunnerSettings(BaseSettings):
    model_config = SettingsConfigDict(frozen=True)

    calibration_library_resolver: ImportString[
        Callable[[Path], QualibrationLibrary]
    ]
    calibration_library_folder: DirectoryPath



from typing import Annotated, Any, Mapping, Type, cast

from fastapi import APIRouter, BackgroundTasks, Depends, HTTPException, status
from pydantic import BaseModel
from qualibrate.qualibration_graph import QualibrationGraph
from qualibrate.qualibration_node import QualibrationNode

from qualibrate_runner.api.dependencies import (
    get_graph as get_qgraph,
)
from qualibrate_runner.api.dependencies import (
    get_node as get_qnode,
)
from qualibrate_runner.api.dependencies import (
    get_state,
)
from qualibrate_runner.config import (
    State,
)
from qualibrate_runner.core.run_job import (
    run_node,
    run_workflow,
    validate_input_parameters,
)

submit_router = APIRouter(prefix="/submit")


@submit_router.post("/node")
def submit_node_run(
    input_parameters: Mapping[str, Any],
    state: Annotated[State, Depends(get_state)],
    node: Annotated[QualibrationNode, Depends(get_qnode)],
    background_tasks: BackgroundTasks,
) -> str:
    # TODO:
    #  this should unify graph submit node params and node submit params
    #  It's needed to correct validation models
    if "parameters" in input_parameters:
        input_parameters = input_parameters["parameters"]
    if state.is_running:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="Already running",
        )
    validate_input_parameters(
        cast(Type[BaseModel], node.parameters_class), input_parameters
    )
    background_tasks.add_task(run_node, node, input_parameters, state)
    return f"Node job {node.name} is submitted"


@submit_router.post("/workflow")
def submit_workflow_run(
    input_parameters: Mapping[str, Any],
    state: Annotated[State, Depends(get_state)],
    graph: Annotated[QualibrationGraph, Depends(get_qgraph)],
    background_tasks: BackgroundTasks,
) -> str:
    if state.is_running:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="Already running",
        )
    input_parameters = {
        "parameters": input_parameters.get("parameters", {}),
        "nodes": {
            name: params.get("parameters", {})
            for name, params in input_parameters.get("nodes", {}).items()
        },
    }
    validate_input_parameters(graph.full_parameters_class, input_parameters)
    background_tasks.add_task(run_workflow, graph, input_parameters, state)
    return f"Workflow job {graph.name} is submitted"



from pathlib import Path
from typing import Any, Mapping, Optional

import click
from pydantic import ValidationError

from qualibrate_runner.config.file import read_config_file
from qualibrate_runner.config.models import QualibrateRunnerSettings
from qualibrate_runner.config.vars import CONFIG_KEY

SUGGEST_MSG = (
    "Can't parse existing config. Fix it or overwrite "
    "by default values using `--overwrite` flag."
)


def get_config_solved_references_or_print_error(
    config_path: Path,
) -> Optional[dict[str, Any]]:
    try:
        return read_config_file(config_path, solve_references=True)
    except ValueError as ex:
        click.secho(str(ex), fg="red")
        click.secho(SUGGEST_MSG, fg="yellow")
    return None


def get_config_model_or_print_error(
    config: Mapping[str, Any],
) -> Optional[QualibrateRunnerSettings]:
    try:
        return QualibrateRunnerSettings(**config)
    except ValidationError as ex:
        errors = [
            (
                f"Message: {error.get('msg')}. "
                "Path: "
                f"{'.'.join([CONFIG_KEY, *map(str, error.get('loc', []))])}. "
                f"Value: {error.get('input')}"
            )
            for error in ex.errors()
        ]
        click.secho("\n".join(errors), fg="red")
        click.secho(SUGGEST_MSG, fg="yellow")
    return None



import traceback
from datetime import datetime
from typing import Any, Mapping, Type

from fastapi import HTTPException, status
from pydantic import BaseModel, ValidationError
from qualibrate.qualibration_graph import QualibrationGraph
from qualibrate.qualibration_library import QualibrationLibrary
from qualibrate.qualibration_node import QualibrationNode

from qualibrate_runner.config import State
from qualibrate_runner.core.models.last_run import (
    LastRun,
    RunError,
    RunnableType,
    RunStatus,
)


def validate_input_parameters(
    parameters_class: Type[BaseModel],
    passed_parameters: Mapping[str, Any],
) -> BaseModel:
    try:
        return parameters_class.model_validate(passed_parameters)
    except ValidationError as ex:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=ex.errors()
        )


def get_active_library_or_error() -> QualibrationLibrary:
    if QualibrationLibrary.active_library is None:
        raise RuntimeError("Qualibration library is not exist")
    return QualibrationLibrary.active_library


def run_node(
    node: QualibrationNode,
    passed_input_parameters: Mapping[str, Any],
    state: State,
) -> None:
    state.run_item = node
    run_status = RunStatus.RUNNING
    state.last_run = LastRun(
        name=node.name,
        status=RunStatus.RUNNING,
        idx=-1,
        passed_parameters=passed_input_parameters,
        started_at=datetime.now(),
        runnable_type=RunnableType.NODE,
    )
    idx = -1
    run_error = None
    try:
        node, _ = node.run(**passed_input_parameters)
    except Exception as ex:
        run_status = RunStatus.ERROR
        run_error = RunError(
            error_class=ex.__class__.__name__,
            message=str(ex),
            traceback=traceback.format_tb(ex.__traceback__),
        )
        run_status = RunStatus.ERROR
        raise
    else:
        idx = node.snapshot_idx if hasattr(node, "snapshot_idx") else -1
        idx = idx if idx is not None else -1
        run_status = RunStatus.FINISHED
    finally:
        state.last_run = LastRun(
            name=state.last_run.name,
            status=run_status,
            idx=idx,
            run_result=node.run_summary,
            runnable_type=state.last_run.runnable_type,
            passed_parameters=passed_input_parameters,
            started_at=state.last_run.started_at,
            completed_at=datetime.now(),
            state_updates=node.state_updates,
            error=run_error,
        )


def run_workflow(
    workflow: QualibrationGraph,
    passed_input_parameters: Mapping[str, Any],
    state: State,
) -> None:
    run_status = RunStatus.RUNNING
    state.last_run = LastRun(
        name=workflow.name,
        status=run_status,
        idx=-1,
        started_at=datetime.now(),
        runnable_type=RunnableType.GRAPH,
        passed_parameters=passed_input_parameters,
    )
    state.run_item = workflow
    idx = -1
    run_error = None
    try:
        library = get_active_library_or_error()
        workflow = library.graphs[workflow.name]
        input_parameters = workflow.full_parameters_class(
            **passed_input_parameters
        )
        workflow.run(
            nodes=input_parameters.nodes.model_dump(),
            **input_parameters.parameters.model_dump(),
        )
    except Exception as ex:
        run_status = RunStatus.ERROR
        run_error = RunError(
            error_class=ex.__class__.__name__,
            message=str(ex),
            traceback=traceback.format_tb(ex.__traceback__),
        )
        raise
    else:
        idx = workflow.snapshot_idx if hasattr(workflow, "snapshot_idx") else -1
        idx = idx if idx is not None else -1
        run_status = RunStatus.FINISHED
    finally:
        state.last_run = LastRun(
            name=state.last_run.name,
            status=run_status,
            idx=idx,
            run_result=workflow.run_summary,
            started_at=state.last_run.started_at,
            completed_at=datetime.now(),
            runnable_type=state.last_run.runnable_type,
            passed_parameters=passed_input_parameters,
            state_updates=(
                workflow.state_updates
                if hasattr(workflow, "state_updates")
                else {}
            ),
            error=run_error,
        )



import sys
from pathlib import Path
from typing import Any, Optional, Union

from qualibrate_runner.config.references.resolvers import resolve_references
from qualibrate_runner.config.vars import (
    DEFAULT_CONFIG_FILENAME,
    DEFAULT_QUALIBRATE_CONFIG_FILENAME,
    QUALIBRATE_PATH,
)

if sys.version_info[:2] < (3, 11):
    import tomli as tomllib
else:
    import tomllib


__all__ = ["get_config_file", "read_config_file"]


def _get_config_file_from_dir(
    dir_path: Path, raise_not_exists: bool = True
) -> Path:
    default_qualibrate = dir_path / DEFAULT_QUALIBRATE_CONFIG_FILENAME
    if default_qualibrate.is_file():
        return default_qualibrate
    default_common = dir_path / DEFAULT_CONFIG_FILENAME
    if default_common.is_file():
        return default_common
    if raise_not_exists:
        raise FileNotFoundError(f"Config file in dir {dir_path} does not exist")
    return default_common


def get_config_file(
    config_path: Optional[Union[str, Path]], raise_not_exists: bool = True
) -> Path:
    if config_path is not None:
        config_path_ = Path(config_path)
        if config_path_.is_file():
            return config_path_
        if config_path_.is_dir():
            return _get_config_file_from_dir(config_path_)
        if raise_not_exists:
            raise OSError("Unexpected config file path")
        return config_path_
    if config_path is None:
        return _get_config_file_from_dir(QUALIBRATE_PATH)


def read_config_file(
    config_file: Path, solve_references: bool = True
) -> dict[str, Any]:
    with config_file.open("rb") as fin:
        config = tomllib.load(fin)
    if not solve_references:
        return config
    return resolve_references(config)
